;*=====================================================================*/
;*    serrano/prgm/project/bigloo/comptime/Cc/ld.scm                   */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Fri Jul 17 09:37:55 1992                          */
;*    Last change :  Tue May 27 07:08:51 2003 (serrano)                */
;*    Copyright   :  1992-2003 Manuel Serrano, see LICENSE file        */
;*    -------------------------------------------------------------    */
;*    The (system) link.                                               */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    Le module                                                        */
;*---------------------------------------------------------------------*/
(module cc_ld
   (export  (ld name ::bool)
	    (lib->string lib . su)
	    (library-suffix)
	    (user-lib-name libname))
   (import  tools_speek
	    tools_error
	    tools_misc
	    cc_exec
	    engine_param
	    engine_configure))

;*---------------------------------------------------------------------*/
;*    ld ...                                                           */
;*---------------------------------------------------------------------*/
(define (ld name need-to-return)
   (cond
      ((string=? (os-class) "unix")
       (unix-ld name need-to-return))
      (else
       (user-error "ld" "Unknown os" (os-class)))))

;*---------------------------------------------------------------------*/
;*    decode-lib-name ...                                              */
;*---------------------------------------------------------------------*/
(define (decode-lib-name lib)
   (match-case lib
      ((? string?)
       (values lib #f))
      (((and ?base (? string?)) . (and ?ver (? string?)))
       (values base ver))
      (else
       (error "ld" "Illegal library name" lib))))

;*---------------------------------------------------------------------*/
;*    encode-lib-name ...                                              */
;*---------------------------------------------------------------------*/
(define (encode-lib-name lib ver)
   (if ver
       (cons lib ver)
       lib))

;*---------------------------------------------------------------------*/
;*    lib->string ...                                                  */
;*---------------------------------------------------------------------*/
(define (lib->string lib . su)
   (multiple-value-bind (base ver)
      (decode-lib-name lib)
      (if ver
	  (string-append base (if (pair? su) (car su) "") "-" ver)
	  (string-append base (if (pair? su) (car su) "")))))

;*---------------------------------------------------------------------*/
;*    lib+suffix ...                                                   */
;*---------------------------------------------------------------------*/
(define (lib+suffix lib static? force?)
   (let* ((n (lib->string lib))
	  (b (if (or (and force? static?)
		     (and (not force?)
			  (or static?
			      *static-bigloo?*
			      (not bgl-configure-shared-library-available?))))
		 (make-static-library-name n)
		 (make-shared-library-name n))))
      (string-append "lib" b)))

;*---------------------------------------------------------------------*/
;*    make-lib-name ...                                                */
;*---------------------------------------------------------------------*/
(define (make-lib-name::bstring lname ss::pair static? force? force-relative?)
   (if (or force-relative?
	   (and *ld-relative* (not (or static? *static-bigloo?*))))
       (string-append "-l" (lib->string lname (car ss)))
       (let loop ((ss ss))
	  (if (null? ss)
	      (error "bigloo"
		     (string-append "Can't find library \""
				    (lib->string lname)
				    "\"")
		     *lib-dir*)
	      (let* ((fname (lib+suffix (lib->string lname (car ss))
					static?
					force?))
		     (name (find-file/path fname *lib-dir*)))
		 (if (string? name)
		     name
		     (begin
			(warning "bigloo"
				 (string-append "Can't find library \""
						fname
						"\"")
				 " in path \"" *lib-dir* "\""
				 (if (pair? (cdr ss))
				     (string-append
				      ".\nTrying replacement library \""
				      (lib+suffix
				       (lib->string lname (cadr ss))
				       static?
				       force?)	
				      "\".")
				     ""))
			(loop (cdr ss)))))))))

;*---------------------------------------------------------------------*/
;*    user-lib-name ...                                                */
;*---------------------------------------------------------------------*/
(define (user-lib-name libname)
   (let ((c (assoc libname *bigloo-libraries-translation-table*)))
      (match-case c
	 (#f libname)
	 ((?- . (? string?)) (cdr c))
	 ((?- . ((? string?) . (? string?))) (cdr c))
	 (else
	  (warning "make-library-name"
		   "Illegal table library name entry -- "
		   c)
	  libname))))

;*---------------------------------------------------------------------*/
;*    library-suffix ...                                               */
;*---------------------------------------------------------------------*/
(define (library-suffix)
   (cond
      (*profile-library* "_p")
      ((>fx *bdb-debug* 1) "_d")
      (*unsafe-library* "_u")
      (else "_s")))

;*---------------------------------------------------------------------*/
;*    secondary-library-suffix ...                                     */
;*---------------------------------------------------------------------*/
(define (secondary-library-suffix)
   (if (>fx *bdb-debug* 1) "_d" ""))
   
;*---------------------------------------------------------------------*/
;*    profile-debug-library-suffix ...                                 */
;*---------------------------------------------------------------------*/
(define (profile-debug-library-suffix)
   (cond
      (*profile-library* "_p")
      ((>fx *bdb-debug* 1) "_d")
      (else "")))
   
;*---------------------------------------------------------------------*/
;*    ld ...                                                           */
;*    -------------------------------------------------------------    */
;*    Le link se fait avec plusieurs fichiers:                         */
;*       1- Le fichier .o resultat de la compilation courrante         */
;*       2- Tous les .o qui ont ete passes en argument                 */
;*       3- Tous les .o qui correspondent aux fichiers de presents     */
;*          dans les clauses `with' du main.                           */
;*---------------------------------------------------------------------*/
(define (unix-ld name need-to-return)
   (verbose 1 "   . ld (" *cc* ")" #\Newline)
   ;; we add additional, machine specific, link options.
   (let ((static? (string-case *ld-options*
		     ((: (* all) "-static")
		      #t)
		     (else
		      #f))))
      (if static?
	  (set! *ld-options* (string-append bgl-configure-static-link-option
					    " " *ld-options*))
	  (set! *ld-options* (string-append bgl-configure-shared-link-option
					    " " *ld-options*)))
      (let* ((dest        (if (string? *dest*)
			      *dest*
			      bgl-configure-a.out))
	     ;; the standard bigloo library
	     (bigloo-lib  (make-lib-name (user-lib-name *bigloo-vlib*)
					 (list (library-suffix)
					       (secondary-library-suffix))
					 static? #f #f))
	     ;; the garbarge collector libary
	     (gc-lib      (make-lib-name (user-lib-name *gc-lib*)
					 (list (profile-debug-library-suffix))
					 (or *profile-library* static?)
					 #f
					 (not *gc-custom?*)))
	     ;; the extra bigloo libraries
	     (add-libs    (let loop ((lib (map user-lib-name
					       *additional-bigloo-libraries*))
				     (res  ""))
			     (if (null? lib)
				 res
				 (loop (cdr lib)
				       (string-append
					(make-lib-name
					 (car lib)
					 (list (library-suffix)
					       (secondary-library-suffix))
					 static?
					 #f
					 #f)
					" "
					res)))))
	     ;; the extra user C libraries
	     (other-libs  (let loop ((lib (reverse *bigloo-user-lib*))
				     (res ""))
			     (if (null? lib)
				 res
				 (loop (cdr lib)
				       (string-append (lib->string (car lib))
						      " " res)))))
	     (ld-args     (string-append
			   ;; object file name
			   name ".o" " "
			   ;; to be linked with files
			   (string*->string *with-files*)
			   ;; other object files
			   (string*->string *o-files*)
			   ;; the executable name
			   " " *ld-o-option* dest
			   ;; cc options
			   " "  *cc-options*
			   ;; optional debug option
			   (if (or *c-debug* (>fx *bdb-debug* 0))
			       (string-append " " *c-debug-option*)
			       "")
			   ;; optional executable stripping
			   (if *strip*
			       (string-append " " bgl-configure-c-strip-flag)
			       "")
			   ;; user ld options
			   " " *ld-options*
			   ;; the library path
			   (let loop ((path *lib-dir*))
			      (if (null? path)
				  ""
				  (string-append "-L"
						 (car path)
						 " "
						 (loop (cdr path)))))
			   ;; additional Bigloo libaries
			   " " add-libs
			   ;; standard bigloo library
			   " " bigloo-lib
			   ;; standard GC library
			   " " gc-lib
			   ;; dloptn library
			   " " bgl-configure-dlopen-lib
			   ;; user libraries
			   " " other-libs
			   ;; then we insert a second time the additional libs
			   " " (if *double-ld-libs?* add-libs "")
			   ;; post user ld options
			   " " *ld-post-options*))
	     (cmd         (string-append *cc* " " ld-args)))
	 (verbose 2 "      ["  cmd #\] #\Newline)
	 (exec cmd need-to-return "ld"))))
