/*=====================================================================*/
/*    serrano/prgm/project/bigloo/fthread/src/BglAsyncThunk.java       */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Mon Feb 11 13:45:20 2002                          */
/*    Last change :  Sat Oct 19 15:20:28 2002 (serrano)                */
/*    Copyright   :  2002 Manuel Serrano                               */
/*    -------------------------------------------------------------    */
/*    A class implementing asynchronous thread. An asynchronous thread */
/*    is implemented by the means of a Java thread that permantenly    */
/*    execute a thunk.                                                 */
/*=====================================================================*/
using System;
using System.Threading;
using bigloo;

namespace bigloo.fairthread
{
  /*---------------------------------------------------------------------*/
  /*    BglAsyncThunk                                                    */
  /*---------------------------------------------------------------------*/
  public class BglAsyncThunk
  {
    private procedure bgl_thunk;
    private static Object async_lock = new Object();
    private static Object async_token = null;
    private Object scheduler;
    private Object _event;

    // The thread associated with this object
    private readonly Thread thread;

    // public constructor
    public BglAsyncThunk( procedure thunk, Object s, Object e, Object id )
    {
      bgl_thunk = thunk;
      thread= new Thread( new ThreadStart( run ) );
      scheduler = s;
      _event = e;
    }

    // lock all the asynchronous thunk
    public static Object funcall( procedure thunk, Object token ) 
    {
      lock( async_lock ) 
      {
        Object res;
	 
        async_token = null;
        res = thunk.funcall1( token );
        async_token = token;
	 
        Monitor.Pulse( async_lock );

        return res;
      }
    }
      
    // lock all the asynchronous thunk
    public static Object funcall2( procedure thunk, Object token, Object val ) 
    {
      lock( async_lock ) 
      {
        Object res;
	 
        async_token = null;
        res = thunk.funcall2( token, val );
        async_token = token;
	 
        Monitor.Pulse( async_lock );

        return res;
      }
    }
      
    // Wait on the global Asynchronous lock
    public static void await( Object token ) 
    {
      try 
      {
        while( async_token != token )
          Monitor.Wait( async_lock );
      } 
      catch( Exception e ) 
      {
        try 
        {
          foreign.internalerror( e );
        } 
        catch( Exception ) 
        {
          Environment.Exit( 1 );
        }
      }
    }

    // display asynchronously a string
    public static Object print( byte[] s, output_port p )
    {
      try
      {
        p.write( s );
        p.flush();
      }
      catch (Exception)
      {
        return bigloo.foreign.BFALSE;
      }
      return s;
    }

    // The thread entry-point
    private void run() 
    {
      try 
      {
        bgl_thunk.funcall2( scheduler, _event );
      } 
      catch( Exception e ) 
      {
        try 
        {
          foreign.internalerror( e );
        } 
        catch( Exception ) 
        {
          Environment.Exit( 1 );
        }
      }
    }

    // Starts the thread
    public void start()
    {
      thread.Start();
    }
  }   
}

