/*=====================================================================*/
/*    .../prgm/project/bigloo/fthread/src/Java/BglAsyncThunk.java      */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Mon Feb 11 13:45:20 2002                          */
/*    Last change :  Fri Jul  4 15:27:51 2003 (serrano)                */
/*    Copyright   :  2002-03 Manuel Serrano                            */
/*    -------------------------------------------------------------    */
/*    A class implementing asynchronous thread. An asynchronous thread */
/*    is implemented by the means of a Java thread that permanently    */
/*    execute a thunk.                                                 */
/*=====================================================================*/

/*---------------------------------------------------------------------*/
/*    The package                                                      */
/*---------------------------------------------------------------------*/
package bigloo.fairthread;
import java.lang.*;
import bigloo.*;

/*---------------------------------------------------------------------*/
/*    BglAsyncThunk                                                    */
/*---------------------------------------------------------------------*/
public class BglAsyncThunk extends Thread {
   private procedure bgl_thunk;
   private static Object async_lock = new Object();
   private static Object async_token = null;
   private Object scheduler;
   private Object event;
   
   // public constructor
   public BglAsyncThunk( procedure thunk, Object s, Object e, Object id ) {
      super();
      bgl_thunk = thunk;
      scheduler = s;
      event = e;
   }

   // lock all the asynchronous thunk
   public static Object funcall( procedure thunk, Object token ) {
      synchronized( async_lock ) {
         Object res;
         
         async_token = null;
         res = thunk.funcall1( token );
         async_token = token;
         
         async_lock.notify();

         return res;
      }
   }
      
   // lock all the asynchronous thunk
   public static Object funcall2( procedure thunk, Object token, Object val ) {
      synchronized( async_lock ) {
         Object res;
         
         async_token = null;
         res = thunk.funcall2( token, val );
         async_token = token;
         
         async_lock.notify();

         return res;
      }
   }
      
    // Wait on the global Asynchronous lock
   public static void await( Object token ) {
      try {
	 while( async_token != token )
	    async_lock.wait();
      } catch( Throwable e ) {
	 try {
	    foreign.internalerror( e );
	 } catch( Throwable _ ) {
	    System.exit( 1 );
	 }
      }
   }

   // display asynchronously a string
   public static Object print( byte[] s, output_port p ) {
      try {
         p.write( s );
         p.flush();
      } catch( Exception _ ) {
	 return bigloo.foreign.BFALSE;
      }
      return s;
   }
   
   // The thread entry-point
   public void run() {
      try {
	 bgl_thunk.funcall2( scheduler, event );
      } catch( Throwable e ) {
	 try {
	    foreign.internalerror( e );
	 } catch( Throwable _ ) {
	    System.exit( 1 );
	 }
      }
   }
}   
