;*=====================================================================*/
;*    serrano/prgm/project/bigloo/fthread/src/Llib/async.scm           */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Sat Oct 19 16:47:03 2002                          */
;*    Last change :  Fri May 23 14:13:24 2003 (serrano)                */
;*    Copyright   :  2002-03 Manuel Serrano                            */
;*    -------------------------------------------------------------    */
;*    Fair thread asyncrent fake signals.                              */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __ft_async

   (import __ft_types
	   __ft_%types
	   __ft_%thread
	   __ft_scheduler
	   __ft_%scheduler
	   __ft_env
	   __ft_%env
	   __ft_signal
	   __ft_reader
	   __ft_writer
	   __ft_socket
	   __ft_process
	   __ft_sleep
	   __ft_exception
	   __ft_%exception
	   __ft_%mutex
	   __ft_thread)

   (export (make-input-signal ::input-port ::obj . opt)
	   (make-output-signal ::output-port ::bstring . opt)
	   (make-connect-signal ::socket . opt)
	   (make-accept-signal ::socket . opt)
	   (make-process-signal ::process . opt)
	   (make-sleep-signal ::long . opt)
	   (make-send-chars-signal ::input-port ::output-port . opt)

	   (fair-read/rp ::procedure ::input-port)))

;*---------------------------------------------------------------------*/
;*    object-equal? ::%lazy-signal ...                                 */
;*---------------------------------------------------------------------*/
(define-method (object-equal? o1::%lazy-signal o2)
   (eq? o1 o2))

;*---------------------------------------------------------------------*/
;*    add-lazy-signal! ...                                             */
;*---------------------------------------------------------------------*/
(define (add-lazy-signal! s::%lazy-signal)
   (with-access::%scheduler (current-scheduler) (signals-tostart)
      (set! signals-tostart (cons s signals-tostart)))
   s)

;*---------------------------------------------------------------------*/
;*    make-input-len-signal ...                                        */
;*---------------------------------------------------------------------*/
(define (make-input-len-signal p::input-port s::int opt)
   (let* ((sl (get-scheduler opt 'make-input-signal))
	  (act (make-fair-reader p s))
	  (e (instantiate::%lazy-signal
		(id 'make-input-signal:len)
		(info p)
		(scheduler sl)
		(action act))))
      (if (input-procedure-port? p)
	  (act sl e)
	  (add-lazy-signal! e))
      e))

;*---------------------------------------------------------------------*/
;*    make-input-pattern-signal ...                                    */
;*---------------------------------------------------------------------*/
(define (make-input-pattern-signal p::input-port s::bstring opt)
   (let* ((sl (get-scheduler opt 'make-input-signal))
	  (act (make-fair-match p (list s)))
	  (e (instantiate::%lazy-signal
		(id 'make-input-signal:pattern)
		(info p)
		(scheduler sl)
		(action act))))
      (if (input-procedure-port? p)
	  (act sl e)
	  (add-lazy-signal! e))
      e))

;*---------------------------------------------------------------------*/
;*    make-input-pattern*-signal ...                                   */
;*---------------------------------------------------------------------*/
(define (make-input-pattern*-signal p::input-port s::pair opt)
   (let* ((sl (get-scheduler opt 'make-input-signal))
	  (act (make-fair-match p s))
	  (e (instantiate::%lazy-signal
		(id 'make-input-signal:pattern)
		(info p)
		(scheduler sl)
		(action act))))
      (if (input-procedure-port? p)
	  (act sl e)
	  (add-lazy-signal! e))
      e))

;*---------------------------------------------------------------------*/
;*    make-input-charset-signal ...                                    */
;*---------------------------------------------------------------------*/
(define (make-input-charset-signal p::input-port s::pair opt)
   (if (pair? (filter (lambda (x) (not (char? x))) s))
       (error "thread-await" "Illegal :input await" opt)
       (let* ((sl (get-scheduler opt 'make-output-signal))
	      (act (make-fair-memq p s))
	      (e (instantiate::%lazy-signal
		    (id 'make-input-signal:charset)
		    (info p)
		    (scheduler sl)
		    (action act))))
	  (if (input-procedure-port? p)
	      (act sl e)
	      (add-lazy-signal! e))
	  e)))

;*---------------------------------------------------------------------*/
;*    make-input-signal ...                                            */
;*---------------------------------------------------------------------*/
(define (make-input-signal p s . opt)
   (define (err)
      (error "make-input-signal" "Illegal argument" s))
   (cond
      ((integer? s)
       (make-input-len-signal p s opt))
      ((string? s)
       (make-input-pattern-signal p s opt))
      ((list? s)
       (cond
	  ((every? char? s)
	   (make-input-charset-signal p s opt))
	  ((every? string? s)
	   (make-input-pattern*-signal p s opt))
	  (else
	   (err))))
      (else
       (err))))

;*---------------------------------------------------------------------*/
;*    make-output-signal ...                                           */
;*---------------------------------------------------------------------*/
(define (make-output-signal p s . opt)
   (let* ((sl (get-scheduler opt 'make-output-signal))
	  (e (instantiate::%lazy-signal
		(id 'make-output-signal)
		(info p)
		(scheduler sl)
		(action (make-fair-writer p s)))))
      (add-lazy-signal! e)
      e))

;*---------------------------------------------------------------------*/
;*    make-connect-signal ...                                          */
;*---------------------------------------------------------------------*/
(define (make-connect-signal s . opt)
   (let* ((sl (get-scheduler opt 'make-connect-signal))
	  (e (instantiate::%lazy-signal
		(id 'make-connect-signal)
		(info s)
		(scheduler sl)
		(action (make-fair-socket s)))))
      (add-lazy-signal! e)
      e))

;*---------------------------------------------------------------------*/
;*    make-accept-signal ...                                           */
;*---------------------------------------------------------------------*/
(define (make-accept-signal s . opt)
   (let* ((sl (get-scheduler opt 'make-accept-signal))
	  (e (instantiate::%lazy-signal
		(id 'make-accept-signal)
		(info s)
		(scheduler sl)
		(action (make-fair-accept-socket s)))))
      (add-lazy-signal! e)
      e))

;*---------------------------------------------------------------------*/
;*    make-process-signal ...                                          */
;*---------------------------------------------------------------------*/
(define (make-process-signal s::process . opt)
   (let* ((sl (get-scheduler opt 'make-process-signal))
	  (e (instantiate::%lazy-signal
		(id 'make-process-signal)
		(info s)
		(scheduler sl)
		(action (make-fair-process s)))))
      (add-lazy-signal! e)
      e))

;*---------------------------------------------------------------------*/
;*    make-sleep-signal ...                                            */
;*---------------------------------------------------------------------*/
(define (make-sleep-signal ms::long . opt)
   (let* ((sl (get-scheduler opt 'make-sleep-signal))
	  (e (instantiate::%lazy-signal
		(id 'make-sleep-signal)
		(info ms)
		(scheduler sl)
		(action (make-fair-sleep ms)))))
      (add-lazy-signal! e)
      e))
		       
;*---------------------------------------------------------------------*/
;*    make-send-chars-signal ...                                       */
;*---------------------------------------------------------------------*/
(define (make-send-chars-signal ip::input-port op::output-port . opt)
   (let* ((len (if (and (pair? opt) (integer? (car opt)))
		   (car opt)
		   -1))
	  (opt (if (and (pair? opt) (integer? (car opt)))
		   (cdr opt)
		   opt))
	  (sl (get-scheduler opt 'make-send-chars-signal))
	  (act (make-fair-send-chars ip op len))
	  (e (instantiate::%lazy-signal
		(id 'make-send-chars-signal)
		(info op)
		(scheduler sl)
		(action act))))
      (if (input-procedure-port? ip)
	  (act sl e)
	  (add-lazy-signal! e))
      e))
		       
;*---------------------------------------------------------------------*/
;*    make-rgc-signal ...                                              */
;*---------------------------------------------------------------------*/
(define (make-rgc-signal state iport def ct)
   (let* ((sl (get-scheduler '() 'make-rgc-signal))
	  (e (instantiate::%lazy-signal
		(id 'make-rgc-signal)
		(info state)
		(scheduler sl)
		(action (lambda (scdl event)
			   (scheduler-broadcast! scdl
						 event
						 (state iport def ct)))))))
      (add-lazy-signal! e)
      e))

;*---------------------------------------------------------------------*/
;*    fair-read/rp ...                                                 */
;*---------------------------------------------------------------------*/
(define (fair-read/rp grammar port)
   (grammar port
	    (lambda (state iport def context)
	       (thread-await! (make-rgc-signal state iport def context)))))

